"""
Workspace module
"""
import os
import sys
import json

import click
import configobj

import onepanel.models.models as models
from onepanel.models.instance import Instance
from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required
from onepanel.commands.projects import ProjectViewController
from onepanel.types.duration import DURATION
from onepanel.types.dataset_mount_identifier import DATASET_MOUNT_IDENTIFIER
from onepanel.types.git_source import GIT_SOURCE


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        if cmd_name == 'ls':
            return click.Group.get_command(self, ctx, 'list')

        return None


class InstanceViewController(APIViewController):
    @staticmethod
    def get_credentials_from_current_path():
        home = os.getcwd()
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        cfg = configobj.ConfigObj(project_file)

        if len(cfg) == 0:
            return None

        return {
            'account_uid': cfg['account_uid'],
            'project_uid': cfg['uid']
        }

    def __init__(self, conn):
        APIViewController.__init__(self, conn)
        self.account_uid = None
        self.uid = None

    def init_credentials_retrieval(self):
        home = os.getcwd()
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        cfg = configobj.ConfigObj(project_file)

        if len(cfg) == 0:
            click.echo("ERROR.Cannot figure out the current project.")
            click.echo("Make sure .onepanel/project exists and is accurate.")
            exit(-1)
        self.account_uid = cfg['account_uid']
        self.uid = cfg['uid']

    def init_endpoint(self):
        self.endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
            root=self.conn.URL,
            account_uid=self.account_uid,
            project_uid=self.uid
        )

    def list(self, account_uid=None, project_uid=None, show_all=False):
        """List all of the workspaces for the provided account_uid, project_uid.
           If show_all is true, and there is no project_uid or account_uid, it will include all workspaces
           for projects the user is a member of too.
        """
        params = None

        if project_uid is not None and account_uid is not None:
            endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
                root=self.conn.URL,
                account_uid=account_uid,
                project_uid=project_uid
            )
        else:
            endpoint = '{root}/instances'.format(
                root=self.conn.URL,
                account_uid=account_uid
            )

        if show_all:
            params = '?show_all=true'

        items = APIViewController.list(self, endpoint=endpoint, params=params)

        if items is None:
            return items

        return [Instance.from_json(item) for item in items]

    def create(self, instance):
        return self.post(post_object=instance, json_encoder=models.APIJSONEncoder)

    def pause(self, account_uid=None, project_uid=None, instance_uid=None):
        if account_uid is None and project_uid is None:
            account_uid = self.account_uid
            project_uid = self.uid

        endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
            root=self.conn.URL,
            account_uid=account_uid,
            project_uid=project_uid
        )

        return self.get('/' + instance_uid, '/stop', endpoint=endpoint)

    def resume(self, account_uid=None, project_uid=None, instance_uid=None):
        if account_uid is None and project_uid is None:
            account_uid = self.account_uid,
            project_uid = self.uid

        endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
            root=self.conn.URL,
            account_uid=account_uid,
            project_uid=project_uid
        )

        return self.get('/' + instance_uid, '/resume', endpoint=endpoint)

    def delete_instance(self, account_uid=None, project_uid=None, instance_uid=None):
        if account_uid is None and project_uid is None:
            account_uid = self.account_uid
            project_uid = self.uid

        endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
            root=self.conn.URL,
            account_uid=account_uid,
            project_uid=project_uid
        )

        return self.delete_v2(instance_uid, endpoint=endpoint)


@click.group(cls=AliasedGroup, help='Workspace commands group')
@click.pass_context
def workspaces(ctx):
    ctx.obj['vc'] = InstanceViewController(ctx.obj['connection'])


@workspaces.command(
    'create',
    help='Create a new workspace. The workspace\'s name: Max 25 chars, lower case alphanumeric or "-", '
         'must start and end with alphanumeric'
)
@click.argument(
    'workspace_uid',
    type=str
)
@click.option(
    '-m', '--machine-type',
    type=str,
    required=True,
    help='Machine type ID. Call "onepanel machine-types list" for IDs.'
)
@click.option(
    '-e', '--environment',
    type=str,
    required=True,
    help='Instance template ID. Call "onepanel environments list" for IDs.'
)
@click.option(
    '-s', '--storage',
    type=str,
    required=True,
    help='Storage type ID.'
)
@click.option(
    '-g', '--source_control',
    type=GIT_SOURCE,
    required=False,
    help='''Source control to use. 

            Format: branch_name/commit_hash

            If omitted, latest commit from master is used.
         '''
)
@click.option(
    '-d', '--mount',
    type=DATASET_MOUNT_IDENTIFIER,
    required=False,
    multiple=True,
    help='''Datasets to mount. 
    
            Format: source=account_name/dataset_name,version=version_of_dataset,destination=folder_to_call_it

            Example: source=onepaneldemo/images,version=3,destination=cats

            The version is optional. If omitted, it'll use the latest version.
            
            The destination is optional. If omitted, it'll use account_name/dataset_name/version
        '''
)
@click.pass_context
@login_required
def create_instance(ctx, workspace_uid, machine_type, environment, storage, mount, source_control):
    try:
        instance = Instance()
        instance.set_uid(workspace_uid)
        instance.machine_type.uid = machine_type
        instance.instance_template.uid = environment
        instance.volume_type.uid = storage
        instance.dataset_mount_claims = mount
        instance.gitSource = source_control

        response = create_instance_internal(ctx.obj['vc'], instance)
        if response['status_code'] == 200:
            click.echo('New workspace created: {}'.format(response['data']['uid']))
        else:
            click.echo('An error occurred with creating a new workspace.')
            click.echo("Details: {}".format(response['data']))

    except ValueError as error:
        click.echo('\tFailed - {}'.format(error))


@workspaces.command(
    'create-from-file',
    help='''Create new workspaces from a json file. 
            The JSON file contains an array. You can have muliple workspaces defined.
        
            \b
            sample.json
            [
                {
                    "name": "test-5",
                    "environment": "jupyter-py3-tensorflow1.11.0",
                    "machine" : "aws-shared",
                    "storage" : "default-storage-10",
                    "git": "master/9ee6aed97073a543b8f8cb56bbab613716f093a1",
                    "datasets": [
                        {
                            "source": "vafilor2/cats",
                            "version": 1,
                            "destination": "cats"
                        }
                    ]  
                }
            ]
            
            Each workspace is created sequentially, top down, one after another. If any fails, they are skipped. 
        '''
)
@click.argument(
    'file_path',
    type=click.Path(exists=True, dir_okay=False)
)
@click.pass_context
@login_required
def create_instances_from_file(ctx, file_path):
    with open(file_path, newline='') as file:
        instances = []

        try:
            instances = json.load(file)
        except json.decoder.JSONDecodeError as error:
            click.echo('Unable to parse json. {}'.format(error))
            return

        for instance in instances:
            click.echo('Creating instance {}'.format(instance['name']))
            try:
                model = Instance.from_simple_json(instance)
                response = create_instance_internal(ctx.obj['vc'], model)
                if response['status_code'] == 200:
                    click.echo('\tSuccess')
                else:
                    click.echo('\tFailed - {}'.format(response['data']))
            except ValueError as error:
                click.echo('\tFailed - {}'.format(error))


def create_instance_internal(vc, instance):
    vc.init_credentials_retrieval()
    vc.init_endpoint()

    return vc.create(instance)


@workspaces.command(
    'list',
    help='''Show workspaces. If this is done outside a project, it will show 
            all of the user's workspaces. If the --all flag is supplied, it will also show workspaces
            for projects the user is a member of.

            If this is done inside a project, only workspaces for that project are listed.
         '''
                    )
@click.option(
    '-a',
    '--show_all',
    is_flag=True,
    default=False,
    help='''If true, shows workspaces for projects the user is a member of.
            This has no effect if you are listing workspaces inside a project.
         '''
              )
@click.option(
    '-q',
    '--id_only',
    is_flag=True,
    default=False,
    help='''Only prints the id information. Inside a project, this is just the workspace uid.
            Outside the project this is a tuple [account_uid, project_uid, workspace_uid]

            No message is printed if no workspaces are found.
         '''
              )
@click.option(
    '-g',
    '--duration_gt',
    type=DURATION,
    help='''Only prints the id information. Inside a project, this is just the workspace uid.
            Outside the project this is account_uid/project_uid/workspace_uid

            No message is printed if no workspaces are found.
         '''
              )
@click.pass_context
@login_required
def list_instances(ctx, show_all, id_only, duration_gt):
    vc = ctx.obj['vc']

    account_uid = None
    project_uid = None

    project_credentials = vc.get_credentials_from_current_path()
    if project_credentials is not None:
        account_uid = project_credentials['account_uid']
        project_uid = project_credentials['project_uid']

    list_instances_internal(vc, account_uid, project_uid, show_all, id_only, duration_gt)


def list_instances_internal(vc, account_uid, project_uid, show_all, id_only, duration_gt=None):
    items = vc.list(account_uid, project_uid, show_all)

    if items is None or len(items) == 0:
        if not id_only:
            click.echo('No workspaces found.')

        return

    if duration_gt is not None:
        if sys.version_info[0] < 3:
            items = filter(lambda item: item.age() > duration_gt, items)
        else:
            items = list(filter(lambda item: item.age() > duration_gt, items))

    formatted_items = [item.simple_view() for item in items]

    project_specific = account_uid is not None and project_uid is not None

    if id_only and not project_specific:
        click.echo(" ".join('{}/{}/{}'.format(item.project.account.uid, item.project.uid, item.uid) for item in items))
    elif id_only and project_specific:
        click.echo(" ".join(item.uid for item in items))
    elif project_specific:
        vc.print_items(formatted_items,
                       fields=['uid', 'state', 'age', 'duration_ready', 'duration_paused', 'cpu', 'gpu', 'ram',
                               'hdd', ],
                       field_names=['ID', 'STATE', 'AGE', 'RUNNING', 'PAUSED', 'CPU', 'GPU', 'RAM', 'HDD'])
    else:
        vc.print_items(formatted_items,
                       fields=['account_uid', 'project_uid', 'uid', 'state', 'age', 'duration_ready', 'duration_paused',
                               'cpu', 'gpu', 'ram', 'hdd', ],
                       field_names=['ACCOUNT', 'PROJECT', 'ID', 'STATE', 'AGE', 'RUNNING', 'PAUSED', 'CPU', 'GPU',
                                    'RAM', 'HDD'])


@workspaces.command(
    'terminate',
    help='''Terminate the workspace(s) given the uids of the workspace
    
            If executing this command inside a project, just the workspace uid is needed.
            Otherwise, you need account_uid/project_uid/workspace_uid
         '''
)
@click.argument(
    'uids',
    type=str,
    nargs=-1
)
@click.pass_context
@login_required
def terminate_instance(ctx, uids):
    vc = ctx.obj['vc']

    if len(uids) == 0:
        click.echo('No uids passed in. No instances deleted')
        return

    for uid in uids:
        response = None

        if vc.get_credentials_from_current_path() is not None:
            vc.init_credentials_retrieval()
            vc.init_endpoint()
            response = vc.delete_instance(instance_uid=uid)
        else:
            ids = uid.split('/')
            response = vc.delete_instance(ids[0], ids[1], ids[2])

        if response['status_code'] == 200:
            click.echo('workspace {} deleted'.format(uid))
        elif response['status_code'] == 404:
            click.echo('workspace {} not found'.format(uid))
        else:
            click.echo('An error occurred with deleting workspace {}.'.format(uid))
            click.echo("Details: {}".format(response['data']))


@workspaces.command(
    'pause',
    help='''Pause the workspace(s) given the uids of the workspace
    
            If executing this command inside a project, just the workspace uid is needed.
            Otherwise, you need account_uid/project_uid/workspace_uid
         '''
)
@click.argument(
    'uids',
    type=str,
    nargs=-1
)
@click.pass_context
@login_required
def pause_instance(ctx, uids):
    vc = ctx.obj['vc']

    if len(uids) == 0:
        click.echo('No uids passed in. No instances paused')
        return

    for uid in uids:
        response = None

        if vc.get_credentials_from_current_path() is not None:
            vc.init_credentials_retrieval()
            vc.init_endpoint()
            response = vc.pause(instance_uid=uid)
        else:
            ids = uid.split('/')
            response = vc.pause(ids[0], ids[1], ids[2])

        if response['status_code'] == 200:
            click.echo('workspace {} paused'.format(uid))
        elif response['status_code'] == 404:
            click.echo('workspace {} not found'.format(uid))
        else:
            click.echo('An error occurred with pausing workspace {}.'.format(uid))
            click.echo("Details: {}".format(response['data']))


@workspaces.command(
    'resume',
    help='''Resume the workspace(s) given the uids of the workspace
    
            If executing this command inside a project, just the workspace uid is needed.
            Otherwise, you need account_uid/project_uid/workspace_uid
         '''
                    )
@click.argument(
    'uids',
    type=str,
    nargs=-1
)
@click.pass_context
@login_required
def resume_instance(ctx, uids):
    vc = ctx.obj['vc']

    if len(uids) == 0:
        click.echo('No uids passed in. No instances resume')
        return

    for uid in uids:
        response = None

        if vc.get_credentials_from_current_path() is not None:
            vc.init_credentials_retrieval()
            vc.init_endpoint()
            response = vc.pause(instance_uid=uid)
        else:
            ids = uid.split('/')
            response = vc.pause(ids[0], ids[1], ids[2])

        if response['status_code'] == 200:
            click.echo('workspace {} resumed'.format(uid))
        elif response['status_code'] == 404:
            click.echo('workspace {} not found'.format(uid))
        else:
            click.echo('An error occurred with resuming workspace {}.'.format(uid))
            click.echo("Details: {}".format(response['data']))



import os

from google.cloud import storage
from onepanel.commands.base import APIViewController


class GCSAuthenticator(APIViewController):
    def __init__(self, conn):
        APIViewController.__init__(self, conn)

    '''
		entity_type is 'datasets' or 'projects'
		entity_uid is 'dataset_uid' or 'project_uid'
	'''

    # todo test this, currently a stub of functionality
    def get_gcs_credentials(self, account_uid='', entity_type='', entity_uid=''):
        if account_uid is '' or entity_type is '' or entity_uid is '':
            return {}  # Empty JSON
        return {"data":"test"}
        # todo implement
        # return self.get('', '/accounts/{account_uid}/{entity_type}/{entity_uid}/credentials/gcs'.format(
        #     account_uid=account_uid,
        #     entity_type=entity_type,
        #     entity_uid=entity_uid
        # ))



import os

from onepanel.utilities.aws_utility import AWSUtility
from onepanel.utilities.gcp_utility import GCPUtility


class CloudStorageUtility:
    @staticmethod
    def get_utility(credentials):
        """
        Use "creds_utility.py" to get the credentials to pass in here.
        :param credentials:
        :return:
        """
        util = None
        cloud_provider = os.getenv('CLOUD_PROVIDER', 'GCP')
        if cloud_provider == 'GCP':
            util = GCPUtility()
        elif cloud_provider == "AWS":
            aws_access_key_id = credentials['data']['AccessKeyID']
            aws_secret_access_key = credentials['data']['SecretAccessKey']
            aws_session_token = credentials['data']['SessionToken']
            util = AWSUtility(aws_access_key_id=aws_access_key_id,
                              aws_secret_access_key=aws_secret_access_key,
                              aws_session_token=aws_session_token)
        if util is None:
            print("Unable to instantiate cloud provider utility. Contact support.")
            exit(-1)
        return util



import json
import platform
import subprocess
import os
import sys
from google.oauth2 import service_account
from google.cloud import storage


class GCPUtility:
    env = {}  # Use for shell command environment variables
    suppress_output = False
    run_cmd_background = False
    # Windows Specific
    # https://docs.microsoft.com/en-us/windows/desktop/ProcThread/process-creation-flags
    # This is used to run processes in the background on Windows
    CREATE_NO_WINDOW = 0x08000000

    def __init__(self):
        json_str = '{"type": "service_account","project_id": "onepanelio","private_key_id": "2eff59c1256e14e88dda2eb97dd9bc0c8f11db96","private_key": "-----BEGIN PRIVATE KEY-----\\nMIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQCfTwpzXYDAnwlM\\nPiNWAk+WwPfufWtIMRfrt1QuJHnukUXwBk0mtLQHleTcI7ny4rO8Dbv5RIbbWOBD\\nv1fj16BgAtUMiDTocgMBHrngWyBUv6DO5IVEdgZGlZFiO672g7cPV9/A2JqFgg1g\\nXA5fA5CKB/JIjNo9ENwqEwp553zGBRKBXCL2FQB7Yy7SzmUipze1oE++KCWSWNUC\\nGF508cB2TT8hRvCs6UCldBsZD152dBmoeEVU2P5C1dgcZfDgqm5YXcbCaq+N/ICW\\n2PMe3073aY3WY8jPf1sN55epA6a57DDfDN2Fv7PcT8HjU5SpcQmNt/CBTjD8/XY9\\nsQciQjQBAgMBAAECggEABigN0ADbCuvvy2HTXRg1xIOskSuI9n8/MhLMTq7bhMav\\nHI7XHcR5+JkTARbjBr86vEJjnfHHVw62NqgHA30VFjYbJE2B40G/WqRkq9uZrOzC\\n5Xvwp9alI3Eq7yWTULTXtuC/JZAOrf4lT4X/NYNzwIHSG8+lrMUhXQGOCWv9Zxbs\\nyp0A24/1PsMyYN75Dj+cxUPzakcJ6kIaePvF9uRBJagaboOMxD2x1Zq58uAshF+K\\nODtrcpTkN46E2Ez2cVUmsoRjDt/VLiQeehfwRYt1Vl2jCVkKHvabuqKVWnlL83vr\\nEwSjD6E/7PgGNvYZGM/+sfE7fX7jS78Vds0oZrNRCQKBgQDVrO3z2+QZu1tmr+QJ\\n2H+9Pozf/q3tI+Kjf8Z03y1INGC0ljIohJxGlD6R7Oqc3S5PHEgC8Lj6UJdX7A42\\n/P8kuIQ87dP6fS0aeuVsM2yS8gMr2njMwuEHPwKc6LkdsWvZQP5KvjX58F0TY26u\\nRfjBPlfMjZ3fg2evj6fJGyhoqwKBgQC+3UdbDFye6TnegHIUaSbWFsaMKW4SS2tF\\nNfJUmCvWEWoSmdcyPo8TNKo5929qAkEmJgALESNwU0XwpBBsgAvIEwmx7rY9Dba9\\njg+5lDXc7F5bWRrzJ26MSXAzUOdMx0jk919twendJH1x4x9RX4S7NbVw99QWcqxc\\n5SmZ0m7uAwKBgHYup43xWUBCOFObhW1MqZpqNlDN9DTtXDVCPun1Zn5BOhun1yZB\\n8o23AMSFXV5Vl396DPpTWnxYhtzCWzpexF/zDnEEufIZuhCjWLF3392Grepbc+UN\\nBJmVaESRE5Iwx43qPLvqaMBNCa3nmRga63M2oCkrSt6NN5QKyKqassbxAoGAKyGi\\nMGcYvyiBVVC9dvNUPWkkMSRjpWq5sQhB3PAOZ+/DgrSVNtuARZzPIw5RdSlRQ6eu\\nttGGqqmG0pygwYQmJW538Ge7cwyYUTm7P/rIOUaJBCecuXodt0IZQt9zhZw7UtK8\\nE/eQO9M1LflfaGoneo7mk8LNgM0lOlDnF8XAH7UCgYEAmeXG+rqA0MLUtucG3xaV\\nkU2ij2LLUeBBEtywi+UF9nRLtrbpU/js8cIN6mXXfkCUfWUUP3ey7q92Eyhx5DDN\\nqZ10H/AVrkluMHuNAUJDctO9GPit32z7PeDuO/AZKbufYgIpqkfShsXZ3kkEIfiC\\nH7r1qakFeFiVwF2Ewm1MbdE=\\n-----END PRIVATE KEY-----\\n","client_email": "gcptesting2@onepanelio.iam.gserviceaccount.com","client_id": "109315516570847019201","auth_uri": "https://accounts.google.com/o/oauth2/auth","token_uri": "https://oauth2.googleapis.com/token","auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs","client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/gcptesting2%40onepanelio.iam.gserviceaccount.com"}'
        if not json_str:
            print("Google credentials cannot be empty! Check the environment variable: {env_var}.".format(env_var='GOOGLE_APPLICATION_CREDENTIALS_JSON_STR'))
            exit(-1)

        service_account_info = json.loads(json_str)
        credentials = service_account.Credentials.from_service_account_info(service_account_info)
        self.storageClient = storage.Client("onepanelio",credentials)
        if platform.system() is 'Windows':
            self.env[str('SYSTEMROOT')] = os.environ['SYSTEMROOT']
            self.env[str('PATH')] = os.environ['PATH']
            self.env[str('PYTHONPATH')] = os.pathsep.join(sys.path)

    def build_full_gcs_url(self, cs_path):
        cs_path = 'gs://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(), path=cs_path)
        return cs_path

    def build_full_cloud_specific_url(self, path):
        return self.build_full_gcs_url(path)

    def get_dataset_bucket_name(self):
        return os.getenv('DATASET_BUCKET', 'onepanel-datasets')

    def upload_dir(self, dataset_directory, gcs_directory, exclude=''):
        bucket = self.storageClient.get_bucket(self.get_dataset_bucket_name())
        dataset_directory_for_upload = dataset_directory.rstrip("/")

        for root, subdirs, files in os.walk(dataset_directory):
            inside_onepanel_dir = False
            file_path_list = root.split(os.path.sep)
            for path_chunk in file_path_list:
                if '.onepanel' == path_chunk:
                    inside_onepanel_dir = True
                    break
            if inside_onepanel_dir:
                continue
            for filename in files:
                root_path_for_gcp = root
                if platform.system() is 'Windows':
                    root_path_for_gcp = root_path_for_gcp.replace("\\","/")
                # We want to set the specified directory as the "current directory" or context of upload.
                # This is to avoid uploading the entire absolute path.
                root_path_for_gcp = root_path_for_gcp.replace(dataset_directory_for_upload,"",1)

                file_path = os.path.join(root, filename)

                # This will be empty for top-level walking, but not sub-directories
                if not root_path_for_gcp:
                    upload_path = "/".join([gcs_directory, filename])
                else:
                    upload_path = "".join([gcs_directory, root_path_for_gcp])
                    upload_path = "/".join([upload_path, filename])
                print("Uploading {file}...".format(file=file_path))
                blob = bucket.blob(upload_path)
                blob.upload_from_filename(file_path)
                print("Uploaded {file}".format(file=file_path))
        return 0

    def download_all(self, dataset_directory, cs_directory):
        """
        This can be run in the background via 'datasets-background-download' CLI command.
        :param dataset_directory: string
        :param cs_directory: string
        :return: int
        """
        bucket = self.storageClient.get_bucket(self.get_dataset_bucket_name())
        blobs = bucket.list_blobs(prefix=cs_directory)
        for blob in blobs:
            blob = bucket.blob(blob.name)
            download_context = blob.name.replace(cs_directory,"",1).lstrip("/")
            if platform.system() is 'Windows':
                download_context = download_context.replace("/", "\\")
            download_path = os.path.sep.join([dataset_directory,download_context])
            # Ensure the parent directories exist before downloading
            if platform.system() is 'Windows':
                separator_before_file = download_path.rfind("\\")
            else:
                separator_before_file = download_path.rfind("/")
            dir_path = download_path[:separator_before_file]
            if not os.path.exists(dir_path):
                os.makedirs(dir_path,exist_ok=True)
            blob.download_to_filename(download_path)
            if self.suppress_output is False:
                print('File {} downloaded to {}.'.format(
                    download_context,
                    download_path))
        return 0

    # TODO temporary method - need to redefine interface for above method
    def download_all_background(self, dataset_directory, s3_directory):
        s3_path = 's3://{bucket}/{path}'.format(bucket=self.get_dataset_bucket_name(), path=s3_directory)
        aws_full_path = self.get_full_path_to_aws_cli()

        if 'aws' not in aws_full_path:
            return -1

        cmd_list = []
        close_fds = False
        if sys.platform != 'win32':
            cmd_list.insert(0, 'nice')
            cmd_list.insert(0, 'nohup')
            close_fds = True
        else:
            # /i so that windows doesn't create "%SYSTEM_DRIVE%" folder
            cmd_list.insert(0, 'start /b /i')
        cmd_list = cmd_list + [aws_full_path, 's3', 'sync', s3_path, dataset_directory]
        # Need to pass the command as one long string. Passing in a list does not work when executed.
        cmd = ' '.join(cmd_list)

        # shell=True because we want to intercept the output from the command.
        # And also, it fixes issues with executing the string of commands.
        if sys.platform != 'win32':
            p = subprocess.Popen(args=cmd, env=self.env, stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
                                 stderr=subprocess.DEVNULL, shell=True, close_fds=close_fds)

            return 0, p
        else:
            p = subprocess.Popen(args=cmd, env=self.env, shell=True, close_fds=close_fds,
                                 creationflags=self.CREATE_NO_WINDOW)

            return 0, p

    # todo support running in the background
    def download(self, to_dir, cloud_provider_full_path_to_file):
        bucket = self.storageClient.get_bucket(self.get_dataset_bucket_name())
        blob = bucket.blob(cloud_provider_full_path_to_file)
        if not blob.exists():
            print("File does not exist.")
            return -1
        file_to_download_idx = blob.name.rfind("/")
        download_context = blob.name[file_to_download_idx:].lstrip("/")
        if platform.system() is 'Windows':
            download_context = download_context.replace("/", "\\")
        download_path = os.path.sep.join([to_dir, download_context])
        blob.download_to_filename(download_path)
        if self.suppress_output is False:
            print('File {} downloaded to {}.'.format(
                download_context,
                download_path))
        return 0

    def check_cloud_path_for_files(self, full_cloud_path='', recursive=True):
        # We have a collision between AWS CLI code and the SDK code of GCP.
        # For AWS, this same function passes in the entire path, such as "s3://<bucket>/<project>/<etc>"
        # The reason for this path is because the AWS CLI needs this notation, so that it can download the files.
        # The GCP SDK does NOT need this, since the bucket object is retrieved from the client, and that
        # already has the "gs://<bucket>" portion of the path.
        # If "gsutil" was compatible with Python 3.x, we could use the same path as for AWS.
        # But it's not, so we have to remove the "gs://<dataset>" prefix, since it'll be passed in.
        path_without_prefix = full_cloud_path.replace(self.build_full_gcs_url(""),"",1)
        if path_without_prefix == '':
            ret_val = {'data': None, 'code': -1, 'msg': 'Need the full cloud path passed in.'}
            return ret_val
        bucket = self.storageClient.get_bucket(self.get_dataset_bucket_name())
        # Another collision between AWS CLI code and the SDK code of GCP.
        # AWSCLI would get summary information, regardless if the path to a file was passed in
        # or a directory.
        # GCP SDK has to do this differently, hence we rely that "recursive" means a directory
        if recursive:
            blobs = bucket.list_blobs(prefix=path_without_prefix)
            # Can't get initialization until you try to iterate through the blobs.
            num_files = 0
            for blob in blobs:
                num_files = num_files + 1
            ret_val = {'data': num_files, 'code': 0, 'msg': 'Total files found.'}
        else:
            blob = bucket.blob(path_without_prefix)
            if blob.exists():
                data = 1
            else:
                data = 0
            ret_val = {'data': data, 'code': 0, 'msg': 'Total files found.'}

        # # todo support recursive
        # recursive_arg = ''
        # if recursive:
        #     recursive_arg = '--recursive'

        return ret_val

    def get_cs_path_details(self, full_cs_path='', total_files=True, total_bytes=True):
        data = {}
        if full_cs_path == '':
            ret_val = {'data': None, 'code': -1, 'msg': 'Need the full google cloud storage path passed in.'}
            return ret_val

        bucket = self.storageClient.get_bucket(self.get_dataset_bucket_name())
        total_files = 0
        total_bytes = 0

        objects_list = bucket.list_blobs(prefix=full_cs_path)
        if objects_list.num_results < 1:
            data['total_bytes'] = total_bytes
            data['total_files'] = total_files
            ret_val = {'data': data, 'code': 0, 'msg': 'Data found.'}
            return ret_val
        else:
            for obj in objects_list:
                total_files += 1
                total_bytes += obj.size

        if total_bytes:
            data['total_bytes'] = total_bytes
        if total_files:
            data['total_files'] = total_files
        ret_val = {'data': data, 'code': 0, 'msg': 'Data found.'}
        return ret_val